from scte.Scte104 import scte104_enums
import copy
import bitstring
import json
byte_size = 8

class SpliceEvent:
    def __init__(self, bitarray_data, init_dict=None):
        if init_dict is not None:
            self.as_dict = init_dict
            return
        message_dict = {}
        message_dict["reserved"] = {"raw": bitarray_data.read("uint:16")}
        message_dict["reserved"]["type"] = scte104_enums.get_op_id_type(message_dict["reserved"]["raw"])
        message_dict["message_size"] = bitarray_data.read("uint:16")
        message_dict["protocol_version"] = bitarray_data.read("uint:8")        
        message_dict["as_index"] = bitarray_data.read("uint:8")
        message_dict["message_number"] = bitarray_data.read("uint:8")
        message_dict["dpi_pid_index"] = bitarray_data.read("uint:16")
        message_dict["scte35_protocol_version"] = bitarray_data.read("uint:8")
        message_dict["timestamp"] = {}
        message_dict["timestamp"]["time_type"] = bitarray_data.read("uint:8")
        
        if message_dict["timestamp"]["time_type"] == 1:
            message_dict["timestamp"]["UTC_seconds"] = bitarray_data.read("uint:32")
            message_dict["timestamp"]["UTC_microseconds"] = bitarray_data.read("uint:16")
        elif message_dict["timestamp"]["time_type"] == 2:
            message_dict["timestamp"]["hours"] = bitarray_data.read("uint:8")
            message_dict["timestamp"]["minutes"] = bitarray_data.read("uint:8")
            message_dict["timestamp"]["seconds"] = bitarray_data.read("uint:8")
            message_dict["timestamp"]["frames"] = bitarray_data.read("uint:8")
        elif message_dict["timestamp"]["time_type"] == 3:
            message_dict["timestamp"]["GP_number"] = bitarray_data.read("uint:8")
            message_dict["timestamp"]["GP_edge"] = bitarray_data.read("uint:8")
        message_dict["num_ops"] = bitarray_data.read("uint:8")
        message_dict["ops"] = []
        for index in range(message_dict["num_ops"]):
            message_dict["ops"].append({})
            message_dict["ops"][index]["op_id"] = bitarray_data.read("uint:16")
            message_dict["ops"][index]["type"] = scte104_enums.get_multi_op_id_type(message_dict["ops"][index]["op_id"])
            message_dict["ops"][index]["data_length"] = bitarray_data.read("uint:16")
            bit_subdata = bitstring.BitString(bytes=bytes.fromhex(bitarray_data.read("hex:" + str(message_dict["ops"][index]["data_length"]*byte_size))))
            message_dict["ops"][index]["data"] = scte104_enums.read_data(message_dict["ops"][index]["op_id"], bit_subdata) 
        self.as_dict = message_dict

    def print(self):
        print(str(self))

    def __str__(self):
        return(json.dumps(self.to_dict(upid_as_str=True), indent=4, sort_keys=False))

    def print_detailed(self):
        print("reserved", hex(self.as_dict['reserved']['raw']), self.as_dict['reserved']['type'])
        print("message_size", hex(self.as_dict['message_size']), self.as_dict['message_size'])
        print("protocol_version", hex(self.as_dict['protocol_version']), self.as_dict['protocol_version'])
        print("as_index", hex(self.as_dict['as_index']), self.as_dict['as_index'])
        print("message_number", hex(self.as_dict['message_number']), self.as_dict['message_number'])
        print("dpi_pid_index", hex(self.as_dict['dpi_pid_index']), self.as_dict['dpi_pid_index'])
        print("scte35_protocol_version", hex(self.as_dict['scte35_protocol_version']), self.as_dict['scte35_protocol_version'])
        print("timestamp", hex(self.as_dict['timestamp']["time_type"]), self.as_dict['timestamp']["time_type"])
        print("num_ops", hex(self.as_dict['num_ops']), self.as_dict['num_ops'])
        for key in self.as_dict['ops']:
            print("op_id", hex(self.as_dict['ops'][key]["op_id"]), self.as_dict['ops'][key]["op_id"], self.as_dict['ops'][key]["type"])
            print("data_length", hex(self.as_dict['ops'][key]["data_length"]), self.as_dict['ops'][key]["data_length"])
            print("data", self.as_dict['ops'][key]["data"])
                
    def to_binary(self):
        self.position = 0
        bit_array = bitstring.BitArray(length=self.as_dict["message_size"]*byte_size)
        self.manipulate_bits(bit_array, self.as_dict["reserved"]["raw"], bytes=2)
        self.manipulate_bits(bit_array, self.as_dict["message_size"], bytes=2)
        self.manipulate_bits(bit_array, self.as_dict["protocol_version"], bytes=1)
        self.manipulate_bits(bit_array, self.as_dict["as_index"], bytes=1)
        self.manipulate_bits(bit_array, self.as_dict["message_number"], bytes=1)
        self.manipulate_bits(bit_array, self.as_dict["dpi_pid_index"], bytes=2)
        self.manipulate_bits(bit_array, self.as_dict["scte35_protocol_version"], bytes=1)
        self.manipulate_bits(bit_array, self.as_dict["timestamp"]["time_type"], bytes=1) ##timestamp

        if self.as_dict["timestamp"]["time_type"] == 1:
            self.manipulate_bits(bit_array, self.as_dict["timestamp"]["UTC_seconds"], bytes=4)
            self.manipulate_bits(bit_array, self.as_dict["timestamp"]["UTC_microseconds"], bytes=2)
        elif self.as_dict["timestamp"]["time_type"] == 2:
            self.manipulate_bits(bit_array, self.as_dict["timestamp"]["hours"], bytes=1)
            self.manipulate_bits(bit_array, self.as_dict["timestamp"]["minutes"], bytes=1)
            self.manipulate_bits(bit_array, self.as_dict["timestamp"]["seconds"], bytes=1)
            self.manipulate_bits(bit_array, self.as_dict["timestamp"]["frames"], bytes=1)
        elif self.as_dict["timestamp"]["time_type"] == 3:
            self.manipulate_bits(bit_array, self.as_dict["timestamp"]["GP_number"], bytes=1)
            self.manipulate_bits(bit_array, self.as_dict["timestamp"]["GP_edge"], bytes=1)

        self.manipulate_bits(bit_array, self.as_dict["num_ops"], bytes=1)
        for index in range(self.as_dict["num_ops"]):
            ## Read in metadatato  bit string
            self.manipulate_bits(bit_array, self.as_dict["ops"][index]["op_id"], bytes=2)
            self.manipulate_bits(bit_array, self.as_dict["ops"][index]["data_length"], bytes=2)
            
            ## read in actual position to metadata
            scte104_enums.encode_data(self.as_dict["ops"][index]["op_id"], bit_array, self.as_dict["ops"][index]["data"], self.position)
        
            ##adjust position by data offset
            self.position = self.position + (self.as_dict["ops"][index]["data_length"]) * byte_size
        return bit_array

    def to_dict(self, upid_as_str=False):
        the_dict = copy.deepcopy(self.as_dict)
        if upid_as_str:
          if "ops" in the_dict:
            for idx,op in enumerate(the_dict['ops']):
              if "data" in op:
                if "segmentation_upid" in op['data']:
                  the_dict["ops"][idx]["data"]['segmentation_upid'] = \
                      str(the_dict["ops"][idx]["data"]['segmentation_upid'])
        return the_dict

    def deep_copy(self):
        return copy.deepcopy(self.as_dict)

    def manipulate_bits(self, bit_array, value, bytes=1):
        hex_val = self.hex_string(value, bytes)
        bit_array.overwrite(hex_val, pos=self.position)
        self.position = self.position + bytes*byte_size
        return None


    def hex_string(self, value, bytes):
        s = hex(value) 
        return '0x' + s[2:].zfill(bytes*2)

    def get_pre_roll_time(self):
        return self.as_dict["ops"][0]["data"]["pre_roll_time"]

    def set_pre_roll_time(self, time):
        self.as_dict["ops"][0]["data"]["pre_roll_time"] = time 



import base64
import copy
import bitstring
from scte.Scte35.TimeSignal import TimeSignal
from scte.Scte35.SpliceDescriptor import SpliceDescriptor
from scte.Scte35.SpliceNull import SpliceNull
from scte.Scte35.SpliceInsert import SpliceInsert
from scte.Scte35.SpliceSchedule import SpliceSchedule
import logging


class SpliceEvent:
    def __parse_splice_descriptors(self, descriptor_loop_length, bitarray_data, init_arr=None, logger=None):
        if logger is not None:
            self._log = logger
        else:
            self._log = logging.getLogger()
        splice_descriptors = []
        bytes_left = descriptor_loop_length
        i=0
        while bytes_left > 0:
            new_descriptor = None
            if init_arr:
                new_descriptor = SpliceDescriptor.from_dict(init_arr[i])
                i += 1
            else:
                new_descriptor = SpliceDescriptor(bitarray_data)
            # Descriptor Length does not include splice_descriptor_tag
            bytes_left -= new_descriptor.as_dict()["descriptor_length"]+2
            splice_descriptors.append(new_descriptor)
        return splice_descriptors

    def __init__(self, b64_data, hex_data=None, init_dict=None):
        super(SpliceEvent, self).__init__()

        if init_dict:
            # On first pass, take all the simple data parameters
            self.splice_info_section = init_dict

            # Now, replace simple dict representations with actual objects
            if init_dict["splice_command_type"] is 0:
                self.is_splice_null = True
                self.splice_info_section["splice_null"] = SpliceNull(None)
            elif init_dict["splice_command_type"] is 4:
                self.is_splice_schedule = True
                self.splice_info_section["splice_schedule"] = SpliceSchedule.from_dict(init_dict["splice_schedule"])
            elif init_dict["splice_command_type"] is 5:
                self.is_splice_insert = True
                self.splice_info_section["splice_insert"] = SpliceInsert.from_dict(init_dict["splice_insert"])
            elif init_dict["splice_command_type"] is 6:
                self.is_time_signal = True
                self.splice_info_section["time_signal"] = TimeSignal.from_dict(init_dict["time_signal"])

            if init_dict["descriptor_loop_length"] > 0:
                self.splice_info_section["splice_descriptors"] = self.__parse_splice_descriptors(self.splice_info_section["descriptor_loop_length"], bitarray_data=None, init_arr=init_dict["splice_descriptors"])

            return

        decoded_data = None
        if b64_data:
            decoded_data = base64.standard_b64decode(b64_data)
        elif hex_data:
            decoded_data = bytes.fromhex(hex_data)
        elif init_dict is None:
            raise TypeError('SpliceEvent must be created with B64 data, Hex data, or a passed-in object')

        bitarray_data = bitstring.BitString(bytes=decoded_data)
        self.splice_info_section = {}
        self.splice_info_section["table_id"] = bitarray_data.read("uint:8")
        self.splice_info_section["section_syntax_indicator"] = bitarray_data.read("bool")
        self.splice_info_section["private"] = bitarray_data.read("bool")
        bitarray_data.pos += 2 # Reserved
        self.splice_info_section["section_length"] = bitarray_data.read("uint:12")
        self.splice_info_section["protocol_version"] = bitarray_data.read("uint:8")
        self.splice_info_section["encrypted_packet"] = bitarray_data.read("bool")
        self.splice_info_section["encryption_algorithm"] = bitarray_data.read("uint:6")
        self.splice_info_section["pts_adjustment"] = bitarray_data.read("uint:33")
        self.splice_info_section["cw_index"] = bitarray_data.read("uint:8")
        self.splice_info_section["tier"] = bitarray_data.read("hex:12")
        self.splice_info_section["splice_command_length"] = bitarray_data.read("uint:12")
        self.splice_info_section["splice_command_type"] = bitarray_data.read("uint:8")

        # the pos pointer of bitarray_data will be mutated after the following constructors
        if self.splice_info_section["splice_command_type"] is 0:
            self.is_splice_null = True
            self.splice_info_section["splice_null"] = SpliceNull(bitarray_data)
        elif self.splice_info_section["splice_command_type"] is 4:
            self.is_splice_schedule = True
            self.splice_info_section["splice_schedule"] = SpliceSchedule(bitarray_data)
        elif self.splice_info_section["splice_command_type"] is 5:
            self.is_splice_insert = True
            self.splice_info_section["splice_insert"] = SpliceInsert(bitarray_data)
        elif self.splice_info_section["splice_command_type"] is 6:
            self.is_time_signal = True
            self.splice_info_section["time_signal"] = TimeSignal(bitarray_data)
        # Loop length is number of total bytes for descriptors
        self.splice_info_section["descriptor_loop_length"] = bitarray_data.read("uint:16")
        if self.splice_info_section["descriptor_loop_length"] > 0:
            self.splice_info_section["splice_descriptors"] = self.__parse_splice_descriptors(self.splice_info_section["descriptor_loop_length"], bitarray_data)

    #
    @property
    def bitstring_format(self):
        """
        Return a formatted string representing the object for use by the bitstring library.

        Note: This function only returns a string for the top-level SpliceEvent
        object. Representative bitstring() functions should be called on all
        constituent objects to make up a full serialized representation.
        """
        bitstring_format = 'uint:8=table_id,' \
                           'bool=section_syntax_indicator,' \
                           'bool=private,' \
                           'uint:2=0,' \
                           'uint:12=section_length,' \
                           'uint:8=protocol_version,' \
                           'bool=encrypted_packet,' \
                           'uint:6=encryption_algorithm,' \
                           'uint:33=pts_adjustment,' \
                           'uint:8=cw_index,' \
                           'hex:12=tier,' \
                           'uint:12=splice_command_length,' \
                           'uint:8=splice_command_type'
        return bitstring_format


    def serialize(self):
        splice_info_section_begin_bs = bitstring.pack(fmt=self.bitstring_format, **self.splice_info_section)

        splice_command_type_bs = None
        if self.splice_info_section["splice_command_type"] is 0:
            raise NotImplementedError('Can not interpret splice_null events')

        elif self.splice_info_section["splice_command_type"] is 4:
            raise NotImplementedError('Can not interpret splice_schedule events')

        elif self.splice_info_section["splice_command_type"] is 5:
            raise NotImplementedError('Can not interpret splice_insert events')

        elif self.splice_info_section["splice_command_type"] is 6:
            splice_command_type_bs = self.splice_info_section["time_signal"].serialize()

        descriptor_loop_length_bs = bitstring.pack(fmt='uint:16=descriptor_loop_length', **self.splice_info_section)

        splice_descriptors_bs = None
        if self.splice_info_section["descriptor_loop_length"] > 0:
            for splice_descriptor in self.splice_info_section["splice_descriptors"]:
                splice_descriptors_bs += splice_descriptor.serialize()

        return splice_info_section_begin_bs + splice_command_type_bs + descriptor_loop_length_bs + splice_descriptors_bs

    @property
    def hex_string(self):
        return self.serialize().hex.upper()

    @property
    def as_dict(self):
        the_dict = copy.deepcopy(self.splice_info_section)
        if "time_signal" in the_dict:
            the_dict["time_signal"] = dict(the_dict["time_signal"])
        if "splice_descriptors" in self.splice_info_section:
            the_dict["splice_descriptors"] = []
            for splice_descriptor in self.splice_info_section["splice_descriptors"]:
                the_dict["splice_descriptors"] += [ splice_descriptor.as_dict(upid_as_str=True) ]
        return the_dict

    @classmethod
    def from_hex_string(cls, hex_string):
        return cls(b64_data=None, hex_data=hex_string)

    @classmethod
    def from_dict(cls, input_dict):
        # Need to do input checking here
        return cls(b64_data=None, hex_data=None, init_dict=input_dict)



import logging


class SpliceInsert:
    def __init__(self, bitarray_data, logger=None):
        if logger is not None:
            self._log = logger
        else:
            self._log = logging.getLogger()
        None



import bitstring
import copy
from scte.Scte35 import SegmentationDescriptor
import logging


class SpliceDescriptor:
    def __init__(self, bitarray_data, init_dict=None, logger=None):
        if logger is not None:
            self._log = logger
        else:
            self._log = logging.getLogger()
        if init_dict:
            if 'reserved1' not in init_dict:
                init_dict['reserved1'] = 127
            if 'reserved2' not in init_dict:
                init_dict['reserved2'] = 31
            self.__obj_dict = init_dict
            return
        new_descriptor = {}
        new_descriptor["splice_descriptor_tag"] = bitarray_data.read("uint:8")
        new_descriptor["descriptor_length"] = bitarray_data.read("uint:8")
        if new_descriptor["splice_descriptor_tag"] not in [0, 1, 2, 3]:
            self._log.info("Skipping Unsupported splice_descriptor_tag: " + str(new_descriptor["splice_descriptor_tag"]))
        new_descriptor["identifier"] = bitarray_data.read("uint:32")
        if new_descriptor["splice_descriptor_tag"] is 0:
            # avail descriptor
            None
        elif new_descriptor["splice_descriptor_tag"] is 1:
            # DTMF Descriptor
            None
        elif new_descriptor["splice_descriptor_tag"] is 2:
            # SegmentationDescriptor
            new_descriptor.update(SegmentationDescriptor(bitarray_data).as_dict)
        elif new_descriptor["splice_descriptor_tag"] is 3:
            # Time Descriptor
            None
        self.__obj_dict = new_descriptor

    @classmethod
    def from_hex_string(cls, hex_string):
        bitarray_data = bitstring.BitString(bytes=bytes.fromhex(hex_string))
        return cls(bitarray_data)

    @classmethod
    def from_dict(cls, input_dict):
        # Need to do input checking here
        return cls(bitarray_data=None, init_dict=input_dict)

    @property
    def bitstring_format(self):
        bitstring_format = 'uint:8=splice_descriptor_tag,' \
                           'uint:8=descriptor_length,' \
                           'uint:32=identifier,' \
                           'uint:32=segmentation_event_id,' \
                           'bool=segmentation_event_cancel_indicator,' \
                           'uint:7=reserved1,'
        if self.__obj_dict['segmentation_event_cancel_indicator'] is False:
            bitstring_format += 'bool=program_segmentation_flag,' \
                                'bool=segmentation_duration_flag,' \
                                'bool=delivery_not_restricted_flag,'
            if self.__obj_dict['delivery_not_restricted_flag'] is False:
                bitstring_format += 'bool=web_delivery_allowed_flag,' \
                                    'bool=no_regional_blackout_flag,' \
                                    'bin:2=device_restrictions,'
            else:
                bitstring_format += 'uint:5=reserved2,'
            if self.__obj_dict['program_segmentation_flag'] is False:
                # Not supported yet
                None
            if self.__obj_dict['segmentation_duration_flag'] is True:
                bitstring_format += 'uint:40=segmentation_duration,'
            bitstring_format += 'uint:8=segmentation_upid_type,' \
                                'uint:8=segmentation_upid_length,' \
                                'bytes:' + str(self.__obj_dict['segmentation_upid_length']) + '=segmentation_upid,' \
                                                                                              'uint:8=segmentation_type_id,' \
                                                                                              'uint:8=segment_num,' \
                                                                                              'uint:8=segments_expected'
            if self.__obj_dict['segmentation_type_id'] in [0x34, 0x36]:
                bitstring_format += ',uint:8=sub_segment_num,' \
                                    'uint:8=sub_segments_expected'
        return bitstring_format

    def serialize(self):
        return bitstring.pack(fmt=self.bitstring_format, **self.__obj_dict)

    @property
    def hex_string(self):
        return self.serialize().hex.upper()

    def as_dict(self, upid_as_str=False):
        the_dict = copy.deepcopy(self.__obj_dict)
        if upid_as_str:
            if "segmentation_upid" in the_dict:
                the_dict['segmentation_upid'] = \
                        str(the_dict['segmentation_upid'])
        return the_dict

    def __str__(self):
        return str(self.as_dict())

    def __repr__(self):
        return self.__str__()

    @classmethod
    def from_dict(cls, input_dict):
        # Need to do input checking here
        return cls(bitarray_data=None, init_dict=input_dict)



from scte.Scte35.SegmentationDescriptor import SegmentationDescriptor
from scte.Scte35.SpliceDescriptor import SpliceDescriptor
from scte.Scte35.SpliceEvent import SpliceEvent
from scte.Scte35.SpliceInsert import SpliceInsert
from scte.Scte35.SpliceNull import SpliceNull
from scte.Scte35.SpliceSchedule import SpliceSchedule
from scte.Scte35.TimeSignal import TimeSignal
from scte.Scte35.scte35_enums import get_message



from scte.Scte104 import read_splice_event


def read_data(op_id, bitarray_data):
    op_data = __op_ids_multi_op[op_id]
    if not op_data: return None

    parse_function = op_data["decode_function"]
    return parse_function(bitarray_data)

def encode_data(op_id, bitarray_data, event_object, position):
    op_data = __op_ids_multi_op[op_id]
    if not op_data: return None

    encode_function = op_data["encode_function"]
    return encode_function(bitarray_data, event_object, position)

def get_op_id_type(op_id):
    if op_id >= 0x0013 and op_id <= 0x00FF: return "RESERVED FOR FUTURE BASIC REQUESTS"
    if op_id >= 0x0100 and op_id <= 0x7FFF: return "Reserved"
    if op_id >= 0x8000 and op_id <= 0xBFFF: return "User Defined"
    if op_id >= 0xC000 and op_id <= 0xFFFE: return "Reserved"
    return __op_ids_single_op[op_id]["type"]


def get_op_id_type_from_hex_string(hex_string):
    return __op_ids_single_op[int(hex_string, 16)]["type"]


def get_multi_op_id_type(op_id):
    if op_id >= 0x0000 and op_id <= 0x00FF: return "Reserved"
    if op_id >= 0x0111 and op_id <= 0x02FF: return "Reserved"
    if op_id >= 0x0302 and op_id <= 0x7FFF: return "Reserved"
    if op_id >= 0x8000 and op_id <= 0xBFFF: return "Reserved"
    if op_id >= 0xC000 and op_id <= 0xFFFE: return "User"
    return __op_ids_multi_op[op_id]["type"]


# TODO: get_multi_message_type is undefined
def get_multi_op_id_type_string(hex_string):
    return None
    #return get_multi_message_type(int(hex_string, 16))


def get_segmentation_type(type_id):
    return __segmentation_type_ids[type_id]["message"]


# TODO: The functions above can't access these objects. Move them to top of file
__op_ids_single_op = {
    0x0000: {
        "type": "general_response_data"
    },
    0x0001: {
        "type": "init_request_data"
    },
    0x0002: {
        "type": "init_response_data"
    },
    0x0003: {
        "type": "alive_request_data"
    },
    0x0004: {
        "type": "alive_response_data"
    },
    0x0005: {
        "type": "User Defined"
    },
    0x0006: {
        "type": "User Defined"
    },
    0x0007: {
        "type": "inject_response_data"
    },
    0x00008: {
        "type": "inject_complete_response_data"
    },
    0x0009: {
        "type": "config_request_data"
    },
    0x000A: {
        "type": "config_response_data"
    },
    0x000B: {
        "type": "provisioning_request_data"
    },
    0x000C: {
        "type": "provisioning_response_data"
    },
    0x000D: {
        "type": "Reserved"
    },
    0x000E: {
        "type": "Reserved"
    },
    0x000F: {
        "type": "fault_request_data"
    },
    0x0010: {
        "type": "fault_response_data"
    },
    0x0011: {
        "type": "AS_alive_request_data"
    },
    0x0012: {
        "type": "AS_alive_response_data"
    },

    0xFFFF: {
        "type": "Reserved for multiple operations"
    }
}

__op_ids_multi_op = {
    0x0100: {
        "type": "inject_section_data_request",
        "decode_function": read_splice_event.inject_section_data_request,
        "encode_function": read_splice_event.inject_section_data_request_encode},
    0x0101: {
        "type": "splice_request_data",
        "decode_function": read_splice_event.splice_request_data,
        "encode_function": read_splice_event.splice_request_data_encode},
    0x0102: {
        "type": "splice_null_request_data",
        "decode_function": read_splice_event.splice_null_request_data,
        "encode_function": read_splice_event.splice_null_request_data_encode},
    0x0103: {
        "type": "start_schedule_download_request_data",
        "decode_function": read_splice_event.start_schedule_download_request_data,
        "encode_function": read_splice_event.start_schedule_download_request_data_encode},
    0x0104: {
        "type": "time_signal_request_data",
        "decode_function": read_splice_event.time_signal_request_data,
        "encode_function": read_splice_event.time_signal_request_data_encode},
    0x0105: {
        "type": "transmit_schedule_request_data",
        "decode_function": read_splice_event.transmit_schedule_request_data,
        "encode_function": read_splice_event.transmit_schedule_request_data_encode},
    0x0106: {
        "type": "component_mode_dpi_request_data",
        "decode_function": read_splice_event.component_mode_dpi_request_data,
        "encode_function": read_splice_event.component_mode_dpi_request_data_encode},
    0x0107: {
        "type": "encrypted_dpi_request_data",
        "decode_function": read_splice_event.encrypted_dpi_request_data,
        "encode_function": read_splice_event.encrypted_dpi_request_data_encode},
    0x0108: {
        "type": "insert_descriptor_request_data",
        "decode_function": read_splice_event.insert_descriptor_request_data,
        "encode_function": read_splice_event.insert_descriptor_request_data_encode},
    0x0109: {
        "type": "insert_DTMF_descriptor_request_data",
        "decode_function": read_splice_event.insert_DTMF_descriptor_request_data,
        "encode_function": read_splice_event.insert_DTMF_descriptor_request_data_encode},
    0x010A: {
        "type": "insert_avail_descriptor_request_data",
        "decode_function": read_splice_event.insert_avail_descriptor_request_data,
        "encode_function": read_splice_event.insert_avail_descriptor_request_data_encode},
    0x010B: {
        "type": "insert_segmentation_descriptor_request_data",
        "decode_function": read_splice_event.insert_segmentation_descriptor_request_data,
        "encode_function": read_splice_event.insert_segmentation_descriptor_request_data_encode},
    0x010C: {
        "type": "proprietary_command_request_data",
        "decode_function": read_splice_event.proprietary_command_request_data,
        "encode_function": read_splice_event.proprietary_command_request_data_encode},
    0x010D: {
        "type": "schedule_component_mode_request_data",
        "decode_function": read_splice_event.schedule_component_mode_request_data,
        "encode_function": read_splice_event.schedule_component_mode_request_data_encode},
    0x010E: {
        "type": "schedule_definition_data",
        "decode_function": read_splice_event.schedule_definition_data,
        "encode_function": read_splice_event.schedule_definition_data_encode},
    0x010F: {
        "type": "insert_tier_data",
        "decode_function": read_splice_event.insert_tier_data,
        "encode_function": read_splice_event.insert_tier_data_encode},
    0x0110: {
        "type": "insert_time_descriptor",
        "decode_function": read_splice_event.insert_time_descriptor,
        "encode_function": read_splice_event.insert_time_descriptor_encode},
    0x0300: {
        "type": "delete_control_word_data",
        "decode_function": read_splice_event.delete_control_word_data,
        "encode_function": read_splice_event.delete_control_word_data_encode},
    0x0301: {
        "type": "update_control_word_data",
        "decode_function": read_splice_event.update_control_word_data,
        "encode_function": read_splice_event.update_control_word_data_encode},
    0xFFFF: {
        "type": "Reserved",
        "decode_function": read_splice_event.reserved,
        "encode_function": read_splice_event.reserved_encode}
}

__segmentation_type_ids = {
    0: {
        "message": "Not Indicated"
    },
    1: {
        "message": "Content Identification"
    },
    16: {
        "message": "Program Start"
    },
    17: {
        "message": "Program End"
    },
    18: {
        "message": "Program Early Termination"
    },
    19: {
        "message": "Program Breakaway"
    },
    20: {
        "message": "Program Resumption"
    },
    21: {
        "message": "Program Runover Planned"
    },
    22: {
        "message": "Program Runover Unplanned"
    },
    23: {
        "message": "Program Overlap Start"
    },
    24: {
        "message": "Program Blackout Override"
    },
    25: {
        "message": "Program Start - In Progress"
    },
    32: {
        "message": "Chapter Start"
    },
    33: {
        "message": "Chapter End"
    },
    34: {
        "message": "Break Start"
    },
    35: {
        "message": "Break End"
    },
    48: {
        "message": "Provider Advertisement Start"
    },
    49: {
        "message": "Provider Advertisement End"
    },
    50: {
        "message": "Distributor Advertisement Start"
    },
    51: {
        "message": "Distributor Advertisement End"
    },
    52: {
        "message": "Provider Placement Opportunity Start"
    },
    53: {
        "message": "Provider Placement Opportunity End"
    },
    54: {
        "message": "Distributor Placement Opportunity Start"
    },
    55: {
        "message": "Distributor Placement Opportunity End"
    },
    64: {
        "message": "Unscheduled Event Start"
    },
    65: {
        "message": "Unscheduled Event End"
    },
    80: {
        "message": "Network Start"
    },
    81: {
        "message": "Network End"
    }
}



# TODO: Unused import
import bitstring
from scte.Scte104 import scte104_enums

def manipulate_bits(bit_array, value, position, bytes):
    hex_val = value
    if type(hex_val) == type(int()):
      hex_val = hex_string(value, bytes)
    bit_array.overwrite(hex_val, pos=position)
    return bytes*8


def hex_string(value, bytes):
    s = hex(value)
    return '0x' + s[2:].zfill(bytes*2)


## Not implemented
def init_request_data(bitarray_data):
    return None

## Not implemented
def init_request_data_encode(bitarray_data, event_object, position):
    return None


## Not implemented
def alive_request_data(bitarray_data):
    return None

def alive_request_data_encode(bitarray_data, event_object, position):
    return None


## Not implemented
def alive_response_data(bitarray_data):
    return None

## Not implemented
def alive_response_data_encode(bitarray_data, event_object, position):
    return None


def splice_request_data(bitarray_data):
    request_data = {}
    request_data["splice_insert_type"] = bitarray_data.read("uint:8")
    request_data["splice_event_id"] = bitarray_data.read("uint:32")
    request_data["unique_program_id"] = bitarray_data.read("uint:16")
    request_data["pre_roll_time"] = bitarray_data.read("uint:16")
    request_data["break_duration"] = bitarray_data.read("uint:16")
    request_data["avail_num"] = bitarray_data.read("uint:8")
    request_data["avails_expected"] = bitarray_data.read("uint:8")
    request_data["auto_return_flag"] = bitarray_data.read("uint:8")
    return request_data

def splice_request_data_encode(bitarray_data, event_object, position):
    position += manipulate_bits(bitarray_data, event_object["splice_insert_type"], position, bytes=1)
    position +=manipulate_bits(bitarray_data, event_object["splice_event_id"], position, bytes=4)
    position +=manipulate_bits(bitarray_data, event_object["unique_program_id"], position, bytes=2)
    position +=manipulate_bits(bitarray_data, event_object["pre_roll_time"], position, bytes=2)
    position +=manipulate_bits(bitarray_data, event_object["break_duration"], position, bytes=2)
    position +=manipulate_bits(bitarray_data, event_object["avail_num"], position, bytes=1)
    position +=manipulate_bits(bitarray_data, event_object["avails_expected"], position, bytes=1)
    position +=manipulate_bits(bitarray_data, event_object["auto_return_flag"], position, bytes=1)
    return None


def  encrypted_dpi_request_data(bitarray_data):
    request_data = {}
    request_data["encryption_algorithm"] = bitarray_data.read("uint:8")
    request_data["cw_index"] = bitarray_data.read("uint:8")
    return request_data

def  encrypted_dpi_request_data_encode(bitarray_data, event_object, position):
    return None


def update_control_word_data(bitarray_data):
    request_data= {}
    request_data["cw_index"] = bitarray_data.read("uint:8") 
    request_data["cw_a"] = bitarray_data.read("uint:64")
    request_data["cw_b"] = bitarray_data.read("uint:64")
    request_data["cw_c"] = bitarray_data.read("uint:64")
    return request_data

def update_control_word_data_encode(bitarray_data, event_object, position):
    return None


def delete_control_word_data(bitarray_data):
    request_data = {}
    request_data["cw_index"] = bitarray_data.read("uint:8")
    return request_data

def delete_control_word_data_encode(bitarray_data, event_object, position):
    return None


## Not implmented
def component_mode_dpi_request_data(bitarray_data):
    return None

## Not implmented
def component_mode_dpi_request_data_encode(bitarray_data, event_object, position):
    return None


## Not implemented
def start_schedule_download_request_data(bitarray_data):
    return None

## Not implemented
def start_schedule_download_request_data_encode(bitarray_data, event_object, position):
    return None

## Not implemented
def schedule_definition_data(bitarray_data):
    return None


## Not implemented
def schedule_definition_data_encode(bitarray_data, event_object, position):
    return None


## Not implemented
def schedule_component_mode_request_data(bitarray_data):
    return None

## Not implemented
def schedule_component_mode_request_data_encode(bitarray_data, event_object, position):
    return None


##Not implemented
def transmit_schedule_request_data(bitarray_data):
    return None


##Not implemented
def transmit_schedule_request_data_encode(bitarray_data, event_object, position):
    return None


def time_signal_request_data(bitarray_data):
    request_data = {}
    request_data["pre_roll_time"] = bitarray_data.read("uint:16")
    return request_data

def time_signal_request_data_encode(bitarray_data, event_object, position):
    position += manipulate_bits(bitarray_data, event_object["pre_roll_time"], position, bytes=2)
    return None


## Not implemented
def insert_avail_descriptor_request_data(bitarray_data):
    return None

## Not implemented
def insert_avail_descriptor_request_data_encode(bitarray_data, event_object, position):
    return None


## Not implemented
def insert_descriptor_request_data(bitarray_data):
    return None

## Not implemented
def insert_descriptor_request_data_encode(bitarray_data, event_object, position):
    return None


def insert_DTMF_descriptor_request_data(bitarray_data):
    request_data = {}
    request_data["pre_roll"] = bitarray_data.read("uint:8")
    request_data["dtmf_length"] = bitarray_data.read("uint:8")
    request_data["dtmf"] = bitarray_data.read("uint:" + str(request_data["dtmf_length"]*8))
    return request_data

def insert_DTMF_descriptor_request_data_encode(bitarray_data, event_object, position):
    position += manipulate_bits(bitarray_data, event_object["pre_roll"], position, bytes=1)
    position += manipulate_bits(bitarray_data, event_object["dtmf_length"], position, bytes=1)
    position += manipulate_bits(bitarray_data, event_object["dtmf"], position, bytes=event_object["dtmf_length"])
    return None


##Not implemented
def proprietary_command_request_data(bitarray_data):
    return None

##Not implemented
def proprietary_command_request_data_encode(bitarray_data, event_object, position):
    return None


def splice_null_request_data(bitarray_data):
    return {}

def splice_null_request_data_encode(bitarray_data, event_object, position):
    return {}


def inject_section_data_request(bitarray_data):
    request_data = {}
    request_data["scte35_command_length"] = bitarray_data.read("uint:16")
    request_data["scte35_protocol_version"] = bitarray_data.read("uint:8")
    request_data["scte35_command_length"] = bitarray_data.read("uint:8")
    request_data["scte35_command_contents"] = bitarray_data.read("uint" + str(request_data["scte35_command_length"]))
    return request_data

def inject_section_data_request_encode(bitarray_data, event_object, position):
    return None


def insert_avail_descriptor_request_data(bitarray_data):
    request_data = {}
    request_data["num_provider_avails"] = bitarray_data.read("uint:8")
    request_data["avails"] = []
    [request_data["avails"].append(bitarray_data.read("uint:32")) for i in range(request_data["num_provider_avails"])]
    return

def insert_avail_descriptor_request_data_encode(bitarray_data, event_object, position):
    return None

def insert_segmentation_descriptor_request_data(bitarray_data):
    request = {}
    
    try:
        request["segmentation_event_id"] = bitarray_data.read("uint:32")
        request["segmentation_event_cancel_indicator"] = bitarray_data.read("uint:8")
        request["duration"] = bitarray_data.read("uint:16")
        request["segmentation_upid_type"] = bitarray_data.read("uint:8")
        request["segmentation_upid_length"] = bitarray_data.read("uint:8")
        request["segmentation_upid"] = bitarray_data.read("bytes:" + str(request["segmentation_upid_length"]))
        request["segmentation_type_id"] = {"decimal": bitarray_data.read("uint:8")}
        request["segmentation_type_id"]["name"] = scte104_enums.get_segmentation_type(request["segmentation_type_id"]["decimal"])
        request["segment_num"] = bitarray_data.read("uint:8")
        request["segments_expected"] = bitarray_data.read("uint:8")
        request["duration_extension_frames"] = bitarray_data.read("uint:8")
        request["delivery_not_restricted_flag"] = bitarray_data.read("uint:8")
        request["web_delivery_allowed_flag"] = bitarray_data.read("uint:8")
        request["no_regional_blackout_flag"] = bitarray_data.read("uint:8")
        request["archive_allowed_flag"] = bitarray_data.read("uint:8")
        request["device_restrictions"] = bitarray_data.read("uint:8")
        request["insert_sub_segment_info"] = bitarray_data.read("uint:8")
        request["sub_segment_num"] = bitarray_data.read("uint:8")
        request["sub_segments_expected"] = bitarray_data.read("uint:8")
    except:
        return request
    return request

def insert_segmentation_descriptor_request_data_encode(bitarray_data, event_object, position):
    position += manipulate_bits(bitarray_data, event_object["segmentation_event_id"], position, bytes=4)
    position += manipulate_bits(bitarray_data, event_object["segmentation_event_cancel_indicator"], position, bytes=1)
    position += manipulate_bits(bitarray_data, event_object["duration"], position, bytes=2)
    position += manipulate_bits(bitarray_data, event_object["segmentation_upid_type"], position, bytes=1)
    position += manipulate_bits(bitarray_data, event_object["segmentation_upid_length"], position, bytes=1)
    position += manipulate_bits(bitarray_data, event_object["segmentation_upid"], position, bytes=event_object["segmentation_upid_length"])
    position += manipulate_bits(bitarray_data, event_object["segmentation_type_id"]["decimal"], position, bytes=1)
    position += manipulate_bits(bitarray_data, event_object["segment_num"], position, bytes=1)
    position += manipulate_bits(bitarray_data, event_object["segments_expected"], position, bytes=1)
    position += manipulate_bits(bitarray_data, event_object["duration_extension_frames"], position, bytes=1)
    position += manipulate_bits(bitarray_data, event_object["delivery_not_restricted_flag"], position, bytes=1)
    position += manipulate_bits(bitarray_data, event_object["web_delivery_allowed_flag"], position, bytes=1)
    position += manipulate_bits(bitarray_data, event_object["no_regional_blackout_flag"], position, bytes=1)
    position += manipulate_bits(bitarray_data, event_object["archive_allowed_flag"], position, bytes=1)
    position += manipulate_bits(bitarray_data, event_object["device_restrictions"], position, bytes=1)
    if "insert_sub_segment_info" in event_object.keys():
        position += manipulate_bits(bitarray_data, event_object["insert_sub_segment_info"], position, bytes=1)
    if "sub_segment_num" in event_object.keys():
        position += manipulate_bits(bitarray_data, event_object["sub_segment_num"], position, bytes=1)
    if "sub_segments_expected" in event_object.keys():
        position += manipulate_bits(bitarray_data, event_object["sub_segments_expected"], position, bytes=1)
    return None

## Not implemented
def proprietary_cmmand_request_data(bitarray_data):
    return None

def proprietary_cmmand_request_data_encode(bitarray_data, event_object, position):
    return None


def insert_tier_data(bitarray_data):
    return {"insert_tier_data": bitarray_data.read("uint:16")}

def insert_tier_data_encode(bitarray_data, event_object, position):
    position += manipulate_bits(bitarray_data, event_object["insert_tier_data"], position, bytes=2)
    return None


## Not implemented
def insert_time_descriptor(bitarray_data):
    return None

## Not implemented
def insert_time_descriptor_encode(bitarray_data, value, position):
    return None


##Not implemented
def reserved(bitarray_data):
    return None

##Not implemented
def reserved_encode(bitarray_data, value, position):
    return None



__segmentation_type_ids = {
    0: {
        "message": "Not Indicated"
    },
    1: {
        "message": "Content Identification"
    },
    16: {
        "message": "Program Start"
    },
    17: {
        "message": "Program End"
    },
    18: {
        "message": "Program Early Termination"
    },
    19: {
        "message": "Program Breakaway"
    },
    20: {
        "message": "Program Resumption"
    },
    21: {
        "message": "Program Runover Planned"
    },
    22: {
        "message": "Program Runover Unplanned"
    },
    23: {
        "message": "Program Overlap Start"
    },
    24: {
        "message": "Program Blackout Override"
    },
    25: {
        "message": "Program Start - In Progress"
    },
    32: {
        "message": "Chapter Start"
    },
    33: {
        "message": "Chapter End"
    },
    34: {
        "message": "Break Start"
    },
    35: {
        "message": "Break End"
    },
    48: {
        "message": "Provider Advertisement Start"
    },
    49: {
        "message": "Provider Advertisement End"
    },
    50: {
        "message": "Distributor Advertisement Start"
    },
    51: {
        "message": "Distributor Advertisement End"
    },
    52: {
        "message": "Provider Placement Opportunity Start"
    },
    53: {
        "message": "Provider Placement Opportunity End"
    },
    54: {
        "message": "Distributor Placement Opportunity Start"
    },
    55: {
        "message": "Distributor Placement Opportunity End"
    },
    64: {
        "message": "Unscheduled Event Start"
    },
    65: {
        "message": "Unscheduled Event End"
    },
    80: {
        "message": "Network Start"
    },
    81: {
        "message": "Network End"
    }
}


def get_message(type_id):
    return __segmentation_type_ids[type_id]["message"]


def get_hex_string(type_id):
    # TODO
    return None


def get_id_from_message(message):
    for type_id in __segmentation_type_ids:
        if __segmentation_type_ids[type_id]["message"] is message:
            return type_id



import bitstring
import logging


class TimeSignal:
    def __init__(self, bitarray_data, init_dict=None, logger=None):
        if logger is not None:
            self._log = logger
        else:
            self._log = logging.getLogger()
        if init_dict:
            self.splice_time = init_dict
            return

        self.splice_time = {}
        self.splice_time["time_specified_flag"] = bitarray_data.read("bool")
        if self.splice_time["time_specified_flag"] is True:
            bitarray_data.pos += 6
            self.splice_time["pts_time"] = bitarray_data.read("uint:33")

    def bitstring_format(self):
        bitstring_format = 'bool=time_specified_flag'

        if self.splice_time["time_specified_flag"] is True:
            bitstring_format += ',' + 'uint:6=0,' \
                                      'uint:33=pts_time'

        return bitstring_format

    def serialize(self):
        return bitstring.pack(fmt=self.bitstring_format(), **self.splice_time)


    def __str__(self):
        return str(self.splice_time)

    def __iter__(self): #overridding this to return tuples of (key,value)
        arr = [('time_specified_flag',self.splice_time['time_specified_flag'])]
        if self.splice_time["time_specified_flag"] is True:
            arr += [('pts_time',self.splice_time['pts_time'])]
        return iter(arr)

    def __repr__(self):
        return self.__str__()

    @classmethod
    def from_dict(cls, input_dict):
        # Need to do input checking here
        return cls(bitarray_data=None, init_dict=input_dict)



from distutils.core import setup
setup(
  name='scte',
  packages=['scte', 'scte.Scte35', 'scte.Scte104'],
  version='1.1.1',
  license='apache-2.0',
  description='Tools for working with SCTE standards.',
  author='James Fining',
  author_email='james.fining@nbcuni.com',
  url='https://github.com/jamesfining/scte',
  download_url='https://github.com/jamesfining/scte/archive/v1.1.1.tar.gz',
  keywords=['scte', 'scte35', 'transport', 'stream', 'broadcast'],
  install_requires=[
          'bitstring'
      ],
  classifiers=[
    'Development Status :: 3 - Alpha',
    'Intended Audience :: Developers',
    'Topic :: Multimedia',
    'License :: OSI Approved :: Apache Software License',
    'Programming Language :: Python :: 3'
  ],
)



from scte.Scte35 import scte35_enums
import logging


class SegmentationDescriptor:
    def __init__(self, bitarray_data, logger=None):
        if logger is not None:
            self._log = logger
        else:
            self._log = logging.getLogger()
        new_descriptor = {}
        new_descriptor["segmentation_event_id"] = bitarray_data.read("uint:32")
        new_descriptor["segmentation_event_cancel_indicator"] = bitarray_data.read("bool")
        # Reserved
        new_descriptor["reserved1"] = bitarray_data.read("uint:7")
        if new_descriptor["segmentation_event_cancel_indicator"] is False:
            new_descriptor["program_segmentation_flag"] = bitarray_data.read("bool")
            new_descriptor["segmentation_duration_flag"] = bitarray_data.read("bool")
            new_descriptor["delivery_not_restricted_flag"] = bitarray_data.read("bool")
            if new_descriptor["delivery_not_restricted_flag"] is False:
                new_descriptor["web_delivery_allowed_flag"] = bitarray_data.read("bool")
                new_descriptor["no_regional_blackout_flag"] = bitarray_data.read("bool")
                new_descriptor["archive_allowed_flag"] = bitarray_data.read("bool")
                new_descriptor["device_restrictions"] = bitarray_data.read("bin:2")
            else:
                # Reserved
                new_descriptor["reserved2"] = bitarray_data.read("uint:5")
            if new_descriptor["program_segmentation_flag"] is False:
                new_descriptor["component_count"] = bitarray_data.read("uint:8")
                loop_counter = new_descriptor["component_count"]
                new_descriptor["components"] = []
                while loop_counter > 0:
                    new_component = {}
                    new_component["component_tag"] = bitarray_data.read("uint:8")
                    # Reserved
                    bitarray_data.pos += 7
                    new_component["pts_offset"] = bitarray_data.read("uint:33")
                    new_descriptor["components"].append(new_component)
            if new_descriptor["segmentation_duration_flag"] is True:
                new_descriptor["segmentation_duration"] = bitarray_data.read("uint:40")
            new_descriptor["segmentation_upid_type"] = bitarray_data.read("uint:8")
            new_descriptor["segmentation_upid_length"] = bitarray_data.read("uint:8")
            # UPID Parsing needs more work
            if new_descriptor["segmentation_upid_length"] > 0:
                new_descriptor["segmentation_upid"] = bitarray_data.read("bytes:" + str(new_descriptor["segmentation_upid_length"]))
            new_descriptor["segmentation_type_id"] = bitarray_data.read("uint:8")
            new_descriptor["segmentation_message"] = scte35_enums.get_message(
                new_descriptor["segmentation_type_id"])
            new_descriptor["segment_num"] = bitarray_data.read("uint:8")
            new_descriptor["segments_expected"] = bitarray_data.read("uint:8")
            if new_descriptor["segmentation_type_id"] in [52, 54]:
                new_descriptor["sub_segment_num"] = bitarray_data.read("uint:8")
                new_descriptor["sub_segments_expected"] = bitarray_data.read("uint:8")
        self.as_dict = new_descriptor
